from __future__ import annotations
import pytest
from .constants import HIGH_NESTING_LEVEL_HTML as HTML
from src.bluemoss import Node, Range, Ex, InvalidXpathException, scrape


def test_valid_xpath():
    Node()
    Node('')
    Node('a')
    Node('/a')
    Node('./a')
    Node('//a')
    Node('.//a')
    Node('///a')


def test_invalid_xpath_1():
    with pytest.raises(InvalidXpathException):
        Node('body///section')


def test_invalid_xpath_2():
    with pytest.raises(InvalidXpathException):
        Node('//')


def test_non_existent_xpath():
    xpath: str = 'h4'
    assert scrape(Node(xpath), HTML) is None
    assert scrape(Node(xpath, filter=3), HTML) is None
    assert scrape(Node(xpath, filter=None), HTML) == []


def test_nested_xpath():
    xpath: str = 'body//section//li/a/../h3'
    assert scrape(Node(xpath, filter=None), HTML) == [
        'Service 1',
        'Service 2',
    ]
    assert scrape(Node(xpath), HTML) == 'Service 1'
    assert scrape(Node(xpath, filter=0), HTML) == 'Service 1'
    assert scrape(Node(xpath, filter=1), HTML) == 'Service 2'
    assert scrape(Node(xpath, filter=Range(0)), HTML) == [
        'Service 1',
        'Service 2',
    ]
    assert scrape(Node(xpath, filter=Range(1)), HTML) == ['Service 2']
    assert scrape(Node(xpath, filter=Range(4)), HTML) == []


def test_nested_xpath_with_tag_attribute_search():
    xpath: str = "body//section//li/a[@href='/info']/../h3"
    assert scrape(Node(xpath, filter=None), HTML) == ['Service 2']
    assert scrape(Node(xpath), HTML) == 'Service 2'


def test_nested_xpath_with_text_search():
    xpath: str = "html//section/ul//h3[text()='Service 1']/../a"
    assert scrape(Node(xpath), HTML) == 'Learn More'
    assert scrape(Node(xpath, extract=Ex.HREF_ENDPOINT), HTML) == '/learn-more'


def test_descendant_axis_flag():
    assert Node().add_descendant_axis_to_xpath is True

    node = Node('h2', add_descendant_axis_to_xpath=False)
    assert scrape(node, HTML) is None

    node.add_descendant_axis_to_xpath = True
    assert scrape(node, HTML) == 'About Us'

    nested_node = Node('main', nodes=[node], add_descendant_axis_to_xpath=False)
    assert scrape(nested_node, HTML) is None

    nested_node.add_descendant_axis_to_xpath = True
    assert scrape(nested_node, HTML) == ['About Us']

    node.add_descendant_axis_to_xpath = False
    assert scrape(nested_node, HTML) == [None]

    node.key = 'key'
    node.add_descendant_axis_to_xpath = True
    assert scrape(nested_node, HTML) == {'key': 'About Us'}

    node.add_descendant_axis_to_xpath = False
    assert scrape(nested_node, HTML) == {'key': None}


def test_xpath_with_preceding_axis():
    assert scrape(Node('//h2'), HTML) == 'About Us'
    assert scrape(Node('./body', extract=Ex.TEXT), HTML) == ''
    assert scrape(Node('/body', extract=Ex.TEXT), HTML) is None


def test_xpath_with_axis_prefix():
    assert Node('div').add_descendant_axis_to_xpath is True
    assert Node('/div').add_descendant_axis_to_xpath is False
    assert Node('./div').add_descendant_axis_to_xpath is False
    assert Node('.//div').add_descendant_axis_to_xpath is False


def test_function_call_xpath():
    assert scrape(Node('count(//a)'), HTML) == 11.0


def test_non_tag_yielding_xpath_with_nodes_1():
    node = Node('count(//a)', nodes=[Node('div'), Node('section')])
    assert scrape(node, HTML) == [None, None]


def test_non_tag_yielding_xpath_with_nodes_2():
    node = Node(
        'div/@class', nodes=[Node('div', key='key_1'), Node('section', key='key_2')]
    )
    assert scrape(node, HTML) == {'key_1': None, 'key_2': None}



import re
from lxml import etree
from functools import cache


@cache
def xpath_is_function_call(xpath: str) -> bool:
    return re.match(r'\w+(-[\w]+)?\(', xpath) is not None


@cache
def xpath_is_valid(xpath_query: str) -> bool:
    """
    Checks if the given @param xpath_query is a valid XPath expression.
    :rtype: bool
    :return: True if @param xpath_query is a valid XPath expression, False otherwise.
    """
    root = etree.Element('root')
    doc = etree.ElementTree(root)
    try:
        doc.xpath(xpath_query)
        return True
    except etree.XPathEvalError:
        return False


__all__ = ['xpath_is_function_call', 'xpath_is_valid']



from __future__ import annotations


def get_infix(text: str, prefix: str, suffix: str) -> str | None:
    """
    Compute the first string-infix in @param text that sits between @param prefix and @param suffix.
    If either one of @param prefix and @param suffix is not found in the text, return None.
    If both @param prefix and @param suffix are present in @param text, the returned string-infix would start
    with the first character after @param prefix ends and ends with the first character before @param suffix starts.

    Example 1: get_infix(text='Hello World!', prefix='ll', suffix='ld') returns 'o Wor'
    Example 2: get_infix(text='Hello World!', prefix='Hey', suffix='World!') returns None
    Example 3: get_infix(text='Hello World!', prefix='Hello', suffix='Darling') returns None

    :param text: String to be searched
    :param prefix: Prefix to be searched
    :param suffix: Suffix to be searched
    :rtype: str | None
    :return: First string-infix in @param text that sits between @param prefix and @param suffix.
    """
    if not text:
        return None
    idx: int = text.find(prefix)
    if idx == -1:
        return None
    res: str = text[idx + len(prefix) :]
    idx = res.find(suffix)
    if idx == -1:
        return None
    return res[:idx]


def clean_text(text: str) -> str:
    """
    Edit the given @param text by
        - removing leading and trailing spaces
        - replacing multiple spaces with just one space
        - replacing any linebreaks + leading space with a linebreak
        - replacing any linebreaks + trailing space with a linebreak
        - replacing multiple newlines with just one newline

    :param text: Text to be edited.
    :rtype: str | None
    :return: The updated text.
    """
    if not text:
        return ''
    text = text.strip()
    while '  ' in text:
        text = text.replace('  ', ' ')
    while '\n ' in text:
        text = text.strip().replace('\n ', '\n')
    while ' \n' in text:
        text = text.strip().replace(' \n', '\n')
    while '\n\n\n' in text:
        text = text.strip().replace('\n\n\n', '\n\n')
    return text


__all__ = ['get_infix', 'clean_text']



from __future__ import annotations
from . import utils
from lxml.html import HtmlElement
from typing import Type, Any, cast
from lxml import html as lxml_html
from .classes import Ex, Node, Range, JsonifyWithTag


def scrape(node: Node, html: str) -> Any:
    """Main function of bluemoss: Scrapes :param html with the scraping-recipe defined by :param node.

    :param node: Node object. Defines how to scrape :param html.
    :param html: HTML doc to be scraped.
    :rtype: Any
    :return: data extracted from :param html
    """
    # noinspection PyTypeChecker
    tag: HtmlElement = lxml_html.fromstring(html)
    return _scrape(node, tag, 0)


def _scrape(node: Node, tag: HtmlElement, level: int) -> Any:
    """Internal main function, which is recursively called in order to scrape :param tag using :param node.

    :param node: Defines how to scrape :param tag.
    :param tag: The html-tag to be scraped.
    :param level:
        The depth level of the current Bluemoss node :param moss (0 at root-level).
        The reason for why we keep track of the depth level of @param moss in relation to it's root node, is because
        we want to allow for the root node (level == 0) to define a value for its parameter 'key' in which case
        we want to return a dict with just one key-value-pair (key pointing to the extracted data).
    :rtype: Any
    :return: data extracted from :param html
    """

    if not isinstance(tag, HtmlElement):
        return None

    if node.no_xpath:
        # if no xpath was defined, we continue to work with the current tag
        tags = [tag]
    else:
        if node.add_descendant_axis_to_xpath:
            xpath = ('//' if level == 0 else './/') + node.xpath
        else:
            xpath = node.xpath
        # find matching tags
        tags = tag.xpath(xpath)
        if not isinstance(tags, list):
            tags = [tags]  # type: ignore[unreachable]
        # filter the matched tags against node.filter
        tags = _filter_tags(node, tags)

    if len(tags) == 0:
        # We either did not find any matching tags or we found matching tags but did not filter any of them.
        return node.transform(None if node.find_single_tag else [])

    if node.target or node.nodes:
        val = [_build_target(node, tag, level) for tag in tags]
    else:
        # len(node.nodes) == 0 and node.target is None
        val = [_scrape_leaf_node(node, tag) for tag in tags]

    if isinstance(node.filter, int):
        # node.filter is an int => we want to filter for a single tag
        assert len(val) == 1
        val = val[0]  # type: ignore

    if level == 0 and node.key is not None:
        # About this if-condition:
        #     1) level == 0: @param moss is most outer parent node (root node)
        #     2) moss.key is not None: extracted data (val) shall be wrapped in a dictionary

        # Why do we only want to get here if @param level is 0?
        #     - If level > 0, then moss.key will be used in the function '_build_target'.
        #     - So, a way to make use of moss.key when level == 0, is to build a dictionary
        #       where moss.key points to @param val.
        return {node.key: node.transform(val)}

    return node.transform(val)


def _filter_tags(node: Node, tags: list[Any]) -> list[Any]:
    """filters the tags matched against node.xpath using node.filter"""

    if isinstance(node.filter, int):
        # node.filter is an int => we want to filter for a single match
        try:
            return [tags[node.filter]]
        except IndexError:
            return []

    # node.filter is not an int => we want to filter for multiple matches

    if isinstance(node.filter, Range):
        # node.filter is a Range object => filter for multiple subsequent matches
        return node.filter.filter(tags)

    if isinstance(node.filter, list):
        # node.filter is a list of integers => filter for the matches whose index is contained in node.filter
        _matches: list[Any] = []
        for idx in node.filter:
            try:
                _matches.append(tags[idx])
            except IndexError:
                _matches.append(None)
        return _matches

    # node.filter is None
    return tags


def _build_target(
    node: Node, tag: HtmlElement | str | None, level: int
) -> list[Any] | dict[str, Any] | Type[Any] | None:
    """
    Builds the target instance (list, dict or class/dataclass instance) of :param node.
    Called by '_extract' function in case :param node defines its 'nodes' parameter.

    :rtype: list | dict | Class (class/dataclass instance)
    :return:
        1. Class instance if the 'target' param of :param node is not None.
        2. List or Dict if the 'target' param of :param node is None:
            2.1: Dict if all nodes in :param node define their 'key' param.
            2.2: List if all nodes have their 'key' param set to None.
    """

    if tag is None:
        return None

    if node.target is None:
        if node.keys_in_nodes:
            # the target is a dict
            return {
                _node.key: _scrape(_node, tag, level + 1)  # type: ignore
                for _node in node.nodes
            }
        # the target is a list
        return [_scrape(_node, tag, level + 1) for _node in node.nodes]

    # the target is a class/dataclass

    # :param values: dictionary to instantiate node.target
    values: dict[str, Any] = {
        _node.key: _scrape(_node, tag, level + 1) for _node in node.nodes  # type: ignore
    }

    if issubclass(node.target, JsonifyWithTag):
        # If node.target inherits from JsonifyWithTag, the intention is to provide
        # the current html-tag @param tag to our target-instance using the key '_tag'.
        values |= {'_tag': tag}

    # noinspection PyArgumentList
    return node.target(**values)  # type: ignore


def _scrape_leaf_node(node: Node, match: Any) -> Any:
    """
    Extracts data from a leaf node.
    A leaf node is a Node object with an empty 'nodes' list (node.nodes == []).

    :rtype: Any
    :return: data extracted from :param node.
    """

    if match is None:
        return None

    if not isinstance(match, HtmlElement):
        # @param tag is a string if node.xpath endswith with e.g. /@href,  /@class or /text(),
        # i.e. when we do not select for a tag but for a tag-property, like text or an attribute.
        return match

    if isinstance(node.extract, str):
        # extract the value of the tag-attribute defined by node.extract
        return match.get(node.extract)

    # extract data from @param tag by match-casing node.extract
    if node.extract == Ex.FULL_TEXT:
        return utils.clean_text(match.text_content().strip())
    elif node.extract == Ex.LXML_HTML_ELEMENT:
        return match
    elif node.extract == Ex.TEXT:
        return match.text.strip()
    elif node.extract == Ex.BS4_TAG:
        return utils.lxml_etree_to_bs4(match)
    elif node.extract == Ex.TAG_AS_STRING:
        return cast(Any, utils.lxml_etree_to_bs4(match)).prettify()
    elif node.extract == Ex.HREF:
        return match.get('href')
    elif node.extract == Ex.HREF_QUERY:
        return utils.get_url_query(match.get('href'))
    elif node.extract == Ex.HREF_DOMAIN:
        return utils.get_domain(match.get('href'))
    elif node.extract == Ex.HREF_ENDPOINT:
        return utils.get_endpoint(match.get('href'))
    elif node.extract == Ex.HREF_BASE_DOMAIN:
        return utils.get_base_domain(match.get('href'))
    elif node.extract == Ex.HREF_QUERY_PARAMS:
        return utils.get_url_query_params(match.get('href'))
    elif node.extract == Ex.HREF_ENDPOINT_WITH_QUERY:
        return utils.get_endpoint_with_query(match.get('href'))
    else:
        raise NotImplementedError


__all__ = ['scrape']



import pytest
from dataclasses import dataclass
from .constants import FOUR_DIVS_HTML as HTML
from src.bluemoss import Node, Range, scrape, EqualIndicesException


def test_find_first():
    expected_output: str = 'Hello 1'
    assert scrape(Node('div'), HTML) == expected_output
    assert scrape(Node('html//div'), HTML) == expected_output


def test_find_last():
    node = Node('div', filter=-1)
    assert scrape(node, HTML) == 'Hello 4'


def test_find_range():
    # first two
    node = Node('div', filter=Range(0, 2))
    assert scrape(node, HTML) == ['Hello 1', 'Hello 2']

    # middle two
    node = Node('div', filter=Range(1, -1))
    assert scrape(node, HTML) == ['Hello 2', 'Hello 3']


def test_find_range_reversed():
    # first two reversed
    node = Node('div', filter=Range(0, 2, reverse=True))
    assert scrape(node, HTML) == ['Hello 2', 'Hello 1']

    # middle two reversed
    node = Node('div', filter=Range(1, -1, True))
    assert scrape(node, HTML) == ['Hello 3', 'Hello 2']


def test_find_all():
    node = Node('div', filter=None)
    assert scrape(node, HTML) == [f'Hello {i}' for i in range(1, 5)]


def test_find_all_reversed():
    node = Node('div', filter=Range(0, None, True))
    assert scrape(node, HTML) == [f'Hello {i}' for i in range(4, 0, -1)]


def test_find_with_list_of_ints():
    node = Node('div', filter=[1, 2, 5, 1, 7])
    assert scrape(node, HTML) == ['Hello 2', 'Hello 3', None, 'Hello 2', None]


def test_find_with_list_of_ints_and_dict_target():
    node = Node('div', filter=[0, 1, 2, 5, 1, 7], nodes=[Node('h1', key='caption')])
    assert scrape(node, HTML) == [
        {'caption': 'Hello 1'},
        {'caption': None},
        {'caption': None},
        None,
        {'caption': None},
        None,
    ]


def test_find_range_with_dict_result():
    # first two
    node = Node(xpath='html', nodes=[Node('div', key='1'), Node('div[2]', key='2')])
    assert scrape(node, HTML) == {'1': 'Hello 1', '2': 'Hello 2'}

    # middle two
    li: list[Node] = [
        Node(
            'html',
            nodes=[
                Node('div', key='1', filter=1),
                Node('div', key='2', filter=2),  # Node("div[3]", key="2")
            ],
        ),
        Node('html', nodes=[Node('div[2]', key='1'), Node('div[3]', key='2')]),
    ]
    for node in li:
        assert scrape(node, HTML) == {'1': 'Hello 2', '2': 'Hello 3'}


def test_find_range_with_class_result():
    @dataclass
    class MyClass:
        a: str
        b: str

        def __eq__(self, other) -> bool:
            return self.a == other.a and self.b == other.b

    # first two
    node = Node(
        'html',
        target=MyClass,
        nodes=[Node('div', key='a'), Node('div[2]', key='b')],
    )
    assert scrape(node, HTML) == MyClass(a='Hello 1', b='Hello 2')

    # middle two
    li: list[Node] = [
        Node(
            'html',
            target=MyClass,
            nodes=[
                Node('div', key='a', filter=1),
                Node('div', key='b', filter=2),
            ],
        ),
        Node(
            'html',
            target=MyClass,
            nodes=[Node('div[2]', key='a'), Node('div[3]', key='b')],
        ),
    ]
    for node in li:
        assert scrape(node, HTML) == MyClass(a='Hello 2', b='Hello 3')


def test_bad_indexing():
    # 1) find many
    node = Node('div', filter=Range(5))
    assert scrape(node, HTML) == []

    # 2) find single & valid (but out of bound) indexing for Range object
    node = Node('div', filter=5)
    assert scrape(node, HTML) is None

    # 3) find single & valid indexing for Range object
    node = Node('div', filter=Range(3, -1))
    assert scrape(node, HTML) == []

    # 4) invalid indexing for Range object
    with pytest.raises(EqualIndicesException):
        Node('div', filter=Range(5, 5))



from __future__ import annotations
from urllib.parse import urlparse, parse_qs


def get_base_domain(url: str | None) -> str | None:
    """
    Returns the base domain of the given @param url.
    Example:
        1) get_base_domain("https://www.google.com?search=python") -> "google.com"
        2) get_base_domain("https://www.products.apple.com?p=iphone") -> "apple.com"
        3) get_base_domain("https://www.sub1.sub2.example.com/about/team?type=engineer") -> "example.com"
        4) get_base_domain(None) -> None

    :rtype: str | None
    :return: The base-domain of the given @param url.
    """
    domain: str | None = get_domain(url)
    return '.'.join(domain.split('.')[-2:]) if domain else None


def get_domain(url: str | None) -> str | None:
    """
    Returns the full domain of the given @param url.
    Example:
        1) get_domain("https://www.google.com?search=python") -> "google.com"
        2) get_domain("https://www.products.apple.com?p=iphone") -> "products.apple.com"
        3) get_domain("https://www.sub1.sub2.example.com/about/team?type=engineer") -> "sub1.sub2.example.com"
        4) get_domain(None) -> None

    :rtype: str | None
    :return: The full domain of @param url.
    """
    url = url.strip() if url else None
    if not url:
        return None
    domain = urlparse(url).netloc
    return domain.replace('www.', '', 1)


def get_endpoint(url: str | None) -> str | None:
    """
    Returns the endpoint @param url.
    Example:
        1) get_endpoint("https://www.google.com?search=python") -> None
        2) get_endpoint("https://www.products.apple.com?p=iphone") -> None
        3) get_endpoint("https://www.sub1.sub2.example.com/about/team?type=engineer") -> "/about/team"
        4) get_endpoint(None) -> None

    :param url: The url to extract the base-domain from.
    :rtype: str | None
    :return: The endpoint of @param url.
    """
    url = url.strip() if url else None
    if not url:
        return None
    endpoint: str | None = urlparse(url).path if url else None
    if endpoint and endpoint.endswith('/'):
        endpoint = endpoint[:-1]
    return endpoint if endpoint else None


def get_url_query(url: str | None) -> str | None:
    """
    Returns the query-string of @param url.
    Example:
        1) get_url_query("https://www.google.com?search=python") -> "search=python"
        2) get_url_query("https://www.products.apple.com?p=iphone") -> "p=iphone"
        3) get_url_query("https://www.sub1.sub2.example.com/about/team?type=engineer") -> "type=engineer"
        4) get_url_query(None) -> None

    :rtype: str | None
    :return: The url-query-string of @param url.
    """
    url = url.strip() if url else None
    if not url:
        return None
    query: str = urlparse(url).query
    return query if query else None


def get_endpoint_with_query(url: str | None) -> str | None:
    """
    Returns the endpoint including query-string of @param url.
    Example:
        1) get_endpoint_with_query("https://www.google.com?search=python") -> "?search=python"
        2) get_endpoint_with_query("https://www.products.apple.com?p=iphone") -> "?p=iphone"
        3) get_endpoint_with_query("https://www.sub1.sub2.example.com/about/team?type=engineer")
                -> "/about/team?type=engineer"
        4) get_endpoint_with_query(None) -> None
        5) get_endpoint_with_query(

    :rtype: str | None
    :return: endpoint + query of @param url.
    """
    if not (endpoint := get_endpoint(url)):
        endpoint = ''
    if not (query := get_url_query(url)):
        query = ''
    if not endpoint and not query:
        return None
    if not query:
        return endpoint
    return f'{endpoint}?{query}'


def get_url_query_params(url: str | None) -> dict[str, list[str]]:
    """
    Returns the query-parameters of @param url as a dict.
    Example:
        1) get_endpoint_with_query("https://www.google.com?search=python") -> {"search": "python"}
        2) get_endpoint_with_query("https://www.products.apple.com?p=iphone") -> {"p": "iphone"}
        3) get_endpoint_with_query("https://www.example.com/team?type=engineer&page=2&type=manager")
                -> {"page": ["2"], "type": ["engineer", "manager"]}
        4) get_endpoint_with_query(None) -> None

    :rtype: dict
    :return: The query-parameters of @param url.
    """
    return parse_qs(urlparse(url).query) if url else {}


__all__ = [
    'get_base_domain',
    'get_domain',
    'get_endpoint',
    'get_url_query',
    'get_endpoint_with_query',
    'get_url_query_params',
]



from __future__ import annotations
import pytest
from datetime import datetime
from dataclasses import dataclass
from .constants import WITH_LINKS_HTML as HTML
from src.bluemoss.utils import url as url_utils
from src.bluemoss.classes.jsonify import Jsonify
from src.bluemoss import (
    Node,
    Ex,
    scrape,
    InvalidTargetTypeException,
    InvalidKeysForTargetException,
    MissingTargetKeysException,
)


@dataclass
class Links(Jsonify):
    links: list[Link]


@dataclass
class Link(Jsonify):
    url: str

    @property
    def domain(self) -> str:
        return url_utils.get_domain(self.url)

    @property
    def base_domain(self) -> str:
        return url_utils.get_base_domain(self.url)

    @property
    def endpoint(self) -> str:
        return url_utils.get_endpoint(self.url)

    @property
    def dict(self) -> dict:
        return {
            'url': self.url,
            'domain': self.domain,
            'endpoint': self.endpoint,
            'base_domain': self.base_domain,
        }


MOSS_WITH_TARGET: Node = Node(
    target=Links,
    nodes=[
        Node(
            xpath='a',
            key='links',
            target=Link,
            filter=None,
            nodes=[Node(key='url', extract=Ex.HREF)],
        )
    ],
)


MOSS_WITHOUT_TARGET: Node = Node(
    xpath='a',
    key='links',
    filter=None,
    nodes=[
        Node(key='url', extract=Ex.HREF),
        Node(key='domain', extract=Ex.HREF_DOMAIN),
        Node(key='base_domain', extract=Ex.HREF_BASE_DOMAIN),
        Node(key='endpoint', extract=Ex.HREF_ENDPOINT),
    ],
)


def test_extract_equality():
    assert scrape(MOSS_WITH_TARGET, HTML).dict == scrape(MOSS_WITHOUT_TARGET, HTML)


def test_target_is_list():
    node = Node('html', nodes=[Node('a', filter=None)])
    assert scrape(node, HTML) == [['Link 1', 'Link 2', 'Link 3', 'Link 4']]


def test_target_is_dict():
    node = Node('html', nodes=[Node('a', key='links', filter=None)])
    assert scrape(node, HTML) == {'links': ['Link 1', 'Link 2', 'Link 3', 'Link 4']}


def test_invalid_target_type():
    with pytest.raises(InvalidTargetTypeException):
        Node(target=list, nodes=[Node(key='some_key')])
    with pytest.raises(InvalidTargetTypeException):
        Node(target=dict, nodes=[Node(key='some_key')])
    with pytest.raises(InvalidTargetTypeException):
        Node(target=set, nodes=[Node(key='some_key')])


def test_invalid_keys_for_target():
    with pytest.raises(InvalidKeysForTargetException):
        Node(target=Link, nodes=[Node(key='domain')])


def test_missing_keys_for_target():
    @dataclass
    class _Link:
        url: str
        created_at: datetime

    expected_message_ending: str = (
        "Missing keys in nodes list for target '_Link': ['created_at']"
    )

    with pytest.raises(MissingTargetKeysException) as exc_info:
        Node(target=_Link, nodes=[Node(key='url')])

    assert str(exc_info.value).endswith(expected_message_ending)



import pytest
from src.bluemoss import Node, PartialKeysException


def test_partial_keys_exception():
    Node(nodes=[])
    Node(nodes=[Node(key='some_key')])
    with pytest.raises(PartialKeysException):
        Node(nodes=[Node(), Node(key='some_key')])



from __future__ import annotations
import abc
from enum import Enum
from typing import Any
from json import dumps
from bs4 import BeautifulSoup
from dataclasses import dataclass
from lxml.html import HtmlElement
from datetime import datetime, date
from collections import OrderedDict
from ..utils import lxml_etree_to_bs4


@dataclass
class Jsonify(abc.ABC):
    """
    An abstract dataclass to provide a .dict and a .json property, in order to turn any dataclass
    instance into a python dict or json object, while
    1. considering the order of parameters as they are defined within each dataclass that inherits from Dictable.
    2. dropping all protected parameters (those which begin with an underscore).

    Consider a dataclass instance p of type Profile(header: Header, pages: list[list[[Page]])
    while Profile, Header and Page are all dataclasses which inherit from Dictable.
    If you now execute .dict on p, the method will not only dictify @param p.header,
    but also all Page instances within the nested list of lists of @param p.pages.
    """

    def __post_init__(self) -> None:
        pass

    def __str__(self) -> str:
        return self.json

    @property
    def json(self) -> str:
        return dumps(self.dict, indent=4)

    @property
    def dict(self) -> OrderedDict[str, Any]:
        d: dict[str, Any] = {
            k: v for k, v in self.__dict__.items() if not k.startswith('_')
        }
        return OrderedDict(
            [
                (key, self.dictify(d[key]))
                for key in self.__dataclass_fields__
                if key in d
            ]
        )

    def dictify(self, val: Any) -> Any:
        if isinstance(val, Enum):
            return val.value
        if isinstance(val, datetime):
            return val.strftime('%Y-%m-%d %H:%M:%S')
        if isinstance(val, date):
            return val.strftime('%Y-%m-%d')
        if isinstance(val, Jsonify):
            return val.dict
        if isinstance(val, list):
            return [self.dictify(v) for v in val]
        if isinstance(val, set):
            return sorted([self.dictify(v) for v in val])
        if isinstance(val, dict):
            return OrderedDict(
                [(self.dictify(k), self.dictify(v)) for k, v in val.items()]
            )
        if hasattr(val, '__dict__'):
            return OrderedDict(
                [(self.dictify(k), self.dictify(v)) for k, v in val.__dict__.items()]
            )
        return val


@dataclass
class JsonifyWithTag(Jsonify):
    """
    Some dataclass instances may need access to their source-html-tag.
    Those dataclasses can inherit from DictableWithTag and thus also get the benefits of the Dictable class.
    """

    _tag: HtmlElement | str

    def __post_init__(self) -> None:
        super().__post_init__()

    @property
    def lxml_tag(self) -> HtmlElement | str:
        return self._tag

    @property
    def bs4_tag(self) -> BeautifulSoup | None:
        return (
            lxml_etree_to_bs4(self._tag) if isinstance(self._tag, HtmlElement) else None
        )

    @property
    def source_line(self) -> int | None:
        """The line within the source-html-doc in which @param self._tag was found."""
        return self._tag.sourceline if isinstance(self._tag, HtmlElement) else None


__all__ = ['Jsonify', 'JsonifyWithTag']



# -*- coding: utf-8 -*-

__title__ = 'bluemoss'
__license__ = 'APACHE'
__version__ = '0.1.30'
__author__ = 'LucaTabone'
__copyright__ = 'Copyright 2023-2023 LucaTabone'

from .classes import *
from .scrape import *



from __future__ import annotations
from lxml import etree
from bs4 import BeautifulSoup
from lxml.html import HtmlElement


def lxml_etree_to_bs4(tag: HtmlElement | str) -> BeautifulSoup | None:
    """
    Transforms a lxml.html.HtmlElement object to a BeautifulSoup object.
    :rtype: BeautifulSoup | None
    :return: BeautifulSoup representation of the given @param tag if isinstance(tag, HTMLElement), None otherwise.
    """
    tag_as_str: str | None = lxml_etree_to_string(tag)
    if tag_as_str is None:
        return None
    return BeautifulSoup(tag_as_str, 'html.parser')


def lxml_etree_to_string(tag: HtmlElement | str) -> str | None:
    """
    Transforms a lxml.html.HtmlElement object to a string.
    :rtype: str | None
    :return: String representation of the given @param tag if isinstance(tag, HTMLElement), None otherwise.
    """
    if not isinstance(tag, HtmlElement):
        return None
    return etree.tostring(tag, method='html').decode('utf-8')


def remove_tags_from_soup(soup: BeautifulSoup, tag_names: list[str]) -> None:
    """
    Removes html-tags with certain ids (@param tag_names) from the given BeautifulSoup object @param soup.
    Exampl: if tag_names is ["h1", "h2"], then all "h1" and "h2" tags will be removed from @param soup.

    :param soup: BeautifulSoup object representing a block of HTML
    :param tag_names: Tag names to be removed HTML
    :return: None
    """
    for data in soup(tag_names):
        data.decompose()


__all__ = ['lxml_etree_to_bs4', 'lxml_etree_to_string', 'remove_tags_from_soup']
